#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
/* Invert ALL channels in a a VSIB data file. Currently this assume
   that the size of and "int" is 4 bytes (ie 32 bits). If not this
   program will fail. It works by XOR'ing the data from every second
   time sample with an appropriate mask for offset binary and
   sign/magnitude data. As it works on 32 bits at a time, it will
   currently not work with mode which use more that a total of 16bits
   of data (or numbers that don't divide into 32
*/

#include <unistd.h>
#include <getopt.h>
#include <values.h>

#define BLOCKSIZE 4*1024*1024

#define RATE 32
#define BIGSTR 1024

typedef enum {VLBA, AT} encoding_type;

int main (int argc, char * const argv[]) {
  int i, infile, outfile, status, nread, nwrote, opt, headsize;
  int tmp;
  unsigned int *data;
  char *mem, msg[256], fname[BIGSTR+1], oname[BIGSTR+1];
  off_t off;
  unsigned int mask, submask;

  encoding_type encoding; /* VLBA or AT bit encoding (offset binary or sign/mag) */
  int bits;         /* Total number of bits/sample (ie nchan*2) */
  int verbose;      /* Prints whats happening */
  int bighead;      /* Use 16 or 816 byte headers */

  struct option options[] = {
    {"bits", 1, 0, 'b'},
    {"bighead", 1, 0, 'r'},
    {"vlba", 0, 0, 'v'},
    {"verbose", 0, 0, 'V'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  if (sizeof(unsigned int)!=4) {
    fprintf(stderr, "This program assumes 32bit word size for the \"int\" type.\n");
    fprintf(stderr, "This does not seem to be the case!! A port to your "
	    "architecture is probably needed!!\n");
    exit(1);
  }

  verbose = 0;
  bits = 8;
  encoding = AT;
  bighead = 0;

  /* Read command line options */
  while (1) {
    opt = getopt_long_only(argc, argv, "", options, NULL);
    if (opt==EOF) break;

    switch (opt) {
      
    case 'b':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -bits option %s\n", optarg);
      else 
	bits = tmp;
      break;

    case 'v':
      encoding = VLBA;
      break;

    case 'r':
      bighead = 1;
      break;

    case 'V':
      verbose = 1;
      break;

    case 'h':
      printf("Usage: band_invert [options] <inputfile> <outputfile>\n");
      printf("  -b/-bits <n>    Total number of bits/sample (total #channels * 2) "
	     "Default 8\n");
      printf("  -v/-vlba        VLBA bit encoding (default AT) (offset binary "
	     "rather than \n");
      printf("                   sign/magnitude\n");
      printf("  -r/-bighead     Asume 816 byte header rather than default 16\n");
      //      printf("  -V/-verbose     Detailed running info\n");
      printf("  -h/-help        This list\n");
      return(1);
      break;

    case '?':
    default:
      break;
    }
  }

  if (argc-optind != 2) {
    fprintf(stderr, "Usange band_invert <inputfile> <outputfile>\n");
    exit(1);
  }

  if (bits>16) {
    fprintf(stderr, "Cannot run with modes using more than a total of 16 bits/sample\n");
    exit(1);
  } else if (bits==1) {
    fprintf(stderr, "Program does not run on 1 bit data!!\n");
  } else if (bits<1) {
    fprintf(stderr, "Meaningless number of bits\n");
  } else if (32 % bits != 0) { /* Number of bits but divide into 32 nicely */
    fprintf(stderr, "Total number of bits must be a divisor of 32 (2,4,8,16)\n");
  }
  
  if (encoding == VLBA) {
    submask = 0x3; /* 11 */
  } else {  /* AT */
    submask = 0x1; /* 01 */ 
  }
  for (i=1; i*2<bits; i++) {
    submask |= (submask&0x3) << i*2;
  }

  mask = 0;
  for (i=0; i*2*bits<32; i++) {
    mask |= submask << i*2*bits;
  }

  mem = malloc(BLOCKSIZE);
  if (mem==NULL) {
    sprintf(msg, "Trying to allocate %d MB", BLOCKSIZE/1024/1024);
    perror(msg);
    return(1);
  }

  strcpy(fname, argv[optind]);
  infile = open(fname, O_RDONLY|O_LARGEFILE);
  if (infile==-1) {
    sprintf(msg, "Failed to open input file (%s)", fname);
    perror(msg);
    free(mem);
    return(1);
  }

  strcpy(oname, argv[optind+1]);
  outfile = open(oname, O_WRONLY|O_CREAT|O_TRUNC|O_LARGEFILE,S_IRWXU);
  if (outfile==-1) {
    sprintf(msg, "Failed to open input file (%s)", oname);
    perror(msg);
    free(mem);
    return(1);
  }

  /* Read the header */
  if (bighead) {
    headsize = 816;
  } else {
    headsize = 16;
  }
  nread = read(infile, mem, headsize ); 

  if (nread==0) {
    sprintf(msg, "\nFailed read header");
    perror(msg);
    close(infile);
    close(outfile);
    return(1);
  } else if (nread != headsize) {
    fprintf(stderr, "\nFailed read header");
    close(infile);
    close(outfile);
    return(1);
  }

  /* Write it out */

  nwrote = write(outfile, mem, headsize);
  if (nwrote==-1) {
    perror("Error writing outfile");
    close(infile);
    close(outfile);
    return(1);
  } else if (nwrote!=headsize) {
    fprintf(stderr, "Warning: Did not write all bytes! (%d/%d)\n",
	    nwrote, headsize);
    close(infile);
    close(outfile);
    return(1);
  }

  while (1) {
    nread = read(infile, mem, BLOCKSIZE);

    if (nread==0) {
      break;
    } else if (nread==-1) {
      perror("Error reading file");
      return(1);
    }

    
    if (! nread%4) { /* Need to work on 32bit words */
      off = lseek(infile, -(nread%4), SEEK_CUR); 
      nread -= nread%4;
    }

    /* Invert the spectrum. Do this by XORing with a mask */
    data = (unsigned int*)mem;
    for (i=0; i<nread; i+=4) {
      *data = *data ^ mask;  
      data++;
    }

    nwrote = write(outfile, mem, nread);

    if (nwrote==-1) {
      perror("Error writing outfile");
      return(1);
    } else if (nwrote!=nread) {
      fprintf(stderr, "Warning: Did not write all bytes! (%d/%d)\n",
	      nwrote, nread);
    } 
  }

  close(outfile);
  close(infile);
  return(1);
}

